/*
server_int.h - interface between engine and the server.dll
Copyright (C) 2013 Uncle Mike

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
*/

#ifndef SERVER_INT_H
#define SERVER_INT_H

#include <alert.h>
#include <cvardef.h>
#include <areanode.h>

#define SV_INTERFACE_VERSION		2

// API errors
#define SV_LIBRARY_LOADED		0	// server.dll was sucessfully loaded
#define SV_TOO_OLD_VERSION		-1	// engine api version it's too old
#define SV_MISMATCH_ENGFUNCS_SIZE	-2	// engine funcs table smaller than server import table 
#define SV_MISMATCH_CALLBACKS_SIZE	-3	// server funcs table smaller than engine expected

// goes into gpGlobals->serverflags
#define SVF_SKIPLOCALHOST		BIT( 0 )	// skip localhost for current player
#define SVF_PLAYERSONLY		BIT( 1 )	// freeze all physics but players
#define SVF_PORTALPASS		BIT( 2 )	// we are do portal pass
#define SVF_APPLY_TRANSITION_LIST	BIT( 3 )	// now entity moves through level, merging changes for global, etc

// 4-22-98  JOHN: added for use in pfnClientPrintf
typedef enum
{
	print_console,
	print_center,
	print_chat,
} PRINT_TYPE;

// for integrity checking of content on clients
typedef enum
{
	force_exactfile,			// File on client must exactly match server's file
	force_model_samebounds,		// For model files only, the geometry must fit in the same bbox
	force_model_specifybounds,		// For model files only, the geometry must fit in the specified bbox
} FORCE_TYPE;

typedef struct cmd_buffer_s
{
	int		argc;		// count of arguments
	const char	*args;		// inline command string
	const char	**argv;		// discrete command string
} cmd_buffer_t;

// save\restore system
#define MAX_CONNECTIONS		16	// These are encoded in the lower 16bits of ENTITYTABLE->flags

// ETABLE->flags
#define FENTTABLE_PLAYER		0x80000000
#define FENTTABLE_REMOVED		0x40000000
#define FENTTABLE_MOVEABLE		0x20000000
#define FENTTABLE_GLOBAL		0x10000000

// TYPEDESCRIPTION->flags
#define FTYPEDESC_GLOBAL		0x0001	// This field is masked for global entity save/restore
#define FTYPEDESC_SAVE		0x0002	// This field is saved to disk
#define FTYPEDESC_KEY		0x0004	// This field can be requested and written to by string name at load time
#define FTYPEDESC_FUNCTIONTABLE	0x0008	// This is a table entry for a member function pointer

typedef enum _fieldtypes
{
	FIELD_VOID = 0,			// No type or value
	FIELD_FLOAT,			// Any floating point value
	FIELD_STRING,			// A string ID (return from ALLOC_STRING)
	FIELD_CLASSPTR,			// CBaseEntity *
	FIELD_EHANDLE,			// Entity handle
	FIELD_EDICT,			// edict_t *, or edict_t *  (same thing)
	FIELD_VECTOR,			// Any vector
	FIELD_POSITION_VECTOR,		// A world coordinate (these are fixed up across level transitions automatically)
	FIELD_INTEGER,			// Any integer or enum
	FIELD_FUNCTION,			// A class function pointer (Think, Use, etc)
	FIELD_BOOLEAN,			// boolean, implemented as an int, I may use this as a hint for compression
	FIELD_SHORT,			// 2 byte integer
	FIELD_CHARACTER,			// a byte
	FIELD_TIME,			// a floating point time (these are fixed up automatically too!)
	FIELD_MODELNAME,			// Engine string that is a model name (needs precache)
	FIELD_SOUNDNAME,			// Engine string that is a sound name (needs precache)
 
	FIELD_TYPECOUNT,			// MUST BE LAST
} FIELDTYPE;

typedef struct
{
	char		mapName[32];
	char		landmarkName[32];
	edict_t		*pentLandmark;
	vec3_t		vecLandmarkOrigin;
} LEVELLIST;

typedef struct 
{
	int		id;		// Ordinal ID of this entity (used for entity <--> pointer conversions)
	edict_t		*pent;		// Pointer to the in-game entity

	int		location;		// Offset from the base data of this entity
	int		size;		// Byte size of this entity's data
	int		flags;		// This could be a short -- bit mask of transitions that this entity is in the PVS of
	string_t		classname;	// entity class name
} ENTITYTABLE;

typedef struct typedesc_s	TYPEDESCRIPTION;	
typedef struct datamap_s	DATAMAP;

typedef struct saverestore_s
{
	char		*pBaseData;		// Start of all entity save data
	char		*pCurrentData;		// Current buffer pointer for sequential access
	int		size;			// Current data size
	int		bufferSize;		// Total space for data
	char		**pTokens;		// Hash table of entity strings (sparse)
	int		tokenCount;		// Number of elements in the pTokens table
	int		tokenSize;		// Size of the linear list of tokens
	ENTITYTABLE	*pTable;			// Array of ENTITYTABLE elements (1 for each entity)
	int		currentIndex;		// Holds a global entity table ID
	int		tableCount;		// Number of elements in the entity table
	LEVELLIST		levelList[MAX_CONNECTIONS];	// List of connections from this level
	int		connectionCount;		// Number of elements in the levelList[]

	// smooth transition
	float		time;
	int		fUseLandmark;
	char		szLandmarkName[32];		// landmark we'll spawn near in next level
	vec3_t		vecLandmarkOffset;		// for landmark transitions
	char		szCurrentMapName[32];	// To check global entities
} SAVERESTOREDATA;

typedef struct sv_enginefuncs_s
{
	size_t		api_size;	// sentinel to compare interface sizes for detect mismatch version	

	// precache specified resource and return unique resource index
	int		(*pfnPrecacheModel)( const char *s );	// preceche model (mdl, bsp, spr)
	int		(*pfnPrecacheSound)( const char *s );	// precache sound or sentence (sentence must leading with '!')
	int		(*pfnPrecacheEvent)( const char *s );	// precache network event
	int		(*pfnPrecacheDecal)( const char *s );	// precache decal

	// tell engine to change level: "mapname", "landmarkname"
	void		(*pfnChangeLevel)( const char *s1, const char *s2 );
	void		(*pfnEndSection)( const char *pszSection ); // trigger_endsection

	// alloc\free edicts
	edict_t*		(*pfnCreateEdict)( int iEntIndex );	// use edict number for reallocate specified edict. Otherwise use -1
	void		(*pfnRemoveEdict)( edict_t *e );	// immediately remove edict from scene

	// memory allocation for entity class data
	void*		(*pfnAllocEntPrivateData)( edict_t *e, size_t cb );

	// given and entity index, returns the corresponding edict pointer
	edict_t*		(*pfnEdictFromIndex)( int iEntIndex );

	// given an edict, returns the entity index
	int		(*pfnIndexFromEdict)( const edict_t *e );

	// valid only for precached models
	int		(*pfnModelIndex)( const char *s );	// index from name
	const char	*(*pfnModelName)( int modelIndex );	// name from index

	// check specified file for consistency
	void		(*pfnForceUnmodified)( FORCE_TYPE type, const float *mins, const float *maxs, const char *filename );

	// move edict onto client, completely remove from server
	void		(*pfnMakeStatic)( edict_t *e );

	// apply static decal to the brush-model (this decal can't moving through levels)
	void		(*pfnStaticDecal)( const float *origin, int decalIndex, int entityIndex, int modelIndex );

	// create some particles for debug purpoces
	void		(*pfnParticleEffect)( const float *org, const float *dir, int color, int count );

	// playback network event
	void		(*pfnPlaybackEvent)( int flags, const edict_t *pInvoker, word ev_index, float delay, const float *org, const float *ang,
			float fparam1, float fparam2, int iparam1, int iparam2, int bparam1, int bparam2 );

	// set global fog	(red<<24|green<<16|blue<<8|density), 0 - disable fog
	void		(*pfnUpdateFogSettings)( unsigned int packed_fog );

	// play regular and ambient sound
	void		(*pfnEmitRegularSound)( edict_t *e, int chan, const char *sample, float vol, float attn, int flags, int pitch );
	void		(*pfnEmitAmbientSound)( edict_t *e, const float *pos, const char *sample, float vol, float attn, int flags, int pitch );
	int		(*pfnBuildSoundMsg)( edict_t *e, const float *p, int chan, const char *s, float vol, float attn, int flags, int pitch );

	// lightstyle management
	const char*	(*pfnGetLightStyle)( int style );
	void		(*pfnSetLightStyle)( int style, const char *value );
	int		(*pfnGetEntityIllum)( const edict_t *e );		// ambient light level of entity

	// file operations
	byte*		(*pfnLoadFile)( const char *filename, long *pLength );
	char*		(*pfnParseFile)( char *data, char *token, int maxlen );
	void		(*pfnFreeFile)( void *buffer );

	// used by node graph code to determine if .bsp and .nod files are out of date
	int		(*pfnCompareFileTime)( const char *filename1, const char *filename2, int *iCompare );
	char**		(*pfnGetFilesList)( const char *pattern, int *numFiles, int gamedironly );	// find in files
	int		(*pfnFileExists)( const char *filename, int gamedironly );

	// random generator
	long		(*pfnRandomLong)( long lLow, long lHigh );
	float		(*pfnRandomFloat)( float flLow, float flHigh );

	// cvar's management (generic functions)
	cvar_t*		(*pfnCvarCreate)( const char *name, const char *value, int flags, const char *description );
	void		(*pfnQueryClientCvarValue)( const edict_t *client, const char *cvarname, int requestID );
	void		(*pfnCvarSet)( cvar_t *var, const char *value );		// same as Cvar_DirectSet
	void		(*pfnCvarReset)( const char *name );			// reset cvar to default value
	cvar_t*		(*pfnCvarFind)( const char *name );			// may return NULL if cvar is not exist

	// command management
	void		(*pfnAddServerCommand)( char *name, void (*function)(void), const char *description );
	cmd_buffer_t*	(*pfnGetCommandBuffer)( void );
	void		(*pfnClientCommand)( const edict_t *client, const char *fmt, ... );
	void		(*pfnServerCommand)( int exec_now, const char *fmt, ... );

	// message handler (NOTE: a some functions works only for local system!)
	void		(*pfnAlertMessage)( ALERT_TYPE type, const char *fmt, ... );// will be formated at game-side
	void		(*pfnClientPrintf)( edict_t *client, PRINT_TYPE ptype, const char *string );
	void		(*pfnServerPrint)( const char *string );		// send message all the clients
	void		(*pfnHost_Error)( const char *errorstring, ... );		// stop the game with error
	int		(*pfnDrawConsoleString)( int x, int y, const char *string );
	void		(*pfnDrawSetTextColor)( float r, float g, float b );
	void		(*pfnDrawConsoleStringLen)( const char *string, int *length, int *height );

	// trace & contents
	int		(*pfnPointContents)( const float *pos, int groupmask );
	trace_t		(*pfnTrace)( const float *p0, const float *mins, const float *maxs, const float *p1, int type, edict_t *e );
	trace_t		(*pfnTraceNoEnts)( const float *p0, const float *mins, const float *maxs, const float *p1, int type, edict_t *e );
	trace_t		(*pfnTraceModel)( const float *v1, const float *mins, const float *maxs, const float *v2, edict_t *pent );
	const char*	(*pfnTraceTexture)( edict_t *pTextureEntity, const float *v1, const float *v2 );

	// physics tools
	double		(*pfnGetServerTime)( void );	// entity think issues
	void		(*pfnSetGroupMask)( int mask, int op );
	void		(*pfnLinkEdict)( edict_t *ent );
	areanode_t*	(*pfnGetHeadnode)( void ); // AABB tree for all physic entities
	void		(*pfnGetBonePosition)( const edict_t *pEdict, int iBone, float *rgflOrigin, float *rgflAngles );
	void		(*pfnGetAttachment)( const edict_t *pEdict, int iAttachment, float *rgflOrigin, float *rgflAngles );
	void*		(*pfnGetModel)( int modelindex );

	// PVS & PAS tools
	byte*		(*pfnSetFatPVS)( const float *org );
	byte*		(*pfnSetFatPAS)( const float *org );
	edict_t*		(*pfnEntitiesInPVS)( const edict_t *pClient );
	int		(*pfnCheckVisibility)( const edict_t *entity, const byte *pset );

	// server info
	int		(*pfnIsDedicatedServer)( void );			// is this a dedicated server?
	int		(*pfnLimitOfEntities)( void );			// maximum edicts count that allowed on server
	int		(*pfnIsMapValid)( const char *filename );		// map is valid for multiplayer?
	void		(*pfnGetGameDirectory)( char *szGetGameDir, int maxlen );

	// client tools
	void		(*pfnGetPlayerStats)( const edict_t *pClient, int *ping, int *packet_loss );
	int		(*pfnGetPlayerUserId)( const edict_t *pClient );		// returns the server assigned userid for this player.
	const char*	(*pfnGetPlayerAuthId)( const edict_t *pClient );		// returns the unique authentification id for this player.
	int		(*pfnCanSkipPlayer)( const edict_t *pClient );		// skip the client updates for weapon prediction
	void		(*pfnFadeClientVolume)( const edict_t *pEdict, int fadePercent, int fadeOutSeconds, int holdTime, int fadeInSeconds );
	edict_t*		(*pfnCreateFakeClient)( const char *netname );		// returns NULL if fake client can't be created
	void		(*pfnRunPlayerMove)( edict_t *fakeclient, const struct usercmd_s *cmd );
	void		(*pfnCrosshairAngle)( const edict_t *pClient, float pitch, float yaw );
	void		(*pfnSetView)( const edict_t *pClient, const edict_t *pViewent );
	edict_t*		(*pfnFindClientInPVS)( const edict_t *pEdict );
	int		(*pfnGetCurrentPlayer)( void );

	// draw collisions etc. Only for local system!
	struct tri_api_s	*pTriAPI;

	// info strings
	char*		(*pfnGetInfoKeyBuffer)( const edict_t *pClient );	// passing in NULL gets the serverinfo
	void		(*pfnSetClientKeyValue)( const edict_t *pClient, char *infobuffer, const char *key, const char *value );
	const char*	(*pfnGetPhysicsKeyValue)( const edict_t *pClient, const char *key );
	void		(*pfnSetPhysicsKeyValue)( const edict_t *pClient, const char *key, const char *value );
	const char*	(*pfnGetPhysicsInfoString)( const edict_t *pClient );
	const char*	(*pfnInfoKeyValue)( const char *infobuffer, const char *key );
	void		(*pfnSetKeyValue)( char *infobuffer, const char *key, const char *value );
	void		(*pfnInfoRemoveKey)( char *infobuffer, const char *key );

	// delta-compression management
	void		(*pfnDeltaAddEncoder)( char *name, void (*pfnEncodeFunc)( struct delta_s *pFields, const byte *from, const byte *to ));
	int		(*pfnDeltaFindField)( struct delta_s *pFields, const char *fieldname );
	void		(*pfnDeltaSetField)	( struct delta_s *pFields, const char *fieldname );
	void		(*pfnDeltaUnsetField)( struct delta_s *pFields, const char *fieldname );
	void		(*pfnDeltaSetFieldByIndex)( struct delta_s *pFields, int fieldNumber );
	void		(*pfnDeltaUnsetFieldByIndex)( struct delta_s *pFields, int fieldNumber );

	// user-message handler
	int		(*pfnRegUserMsg)( const char *pszName, int iSize );
	void		(*pfnMessageBegin)( int msg_dest, int msg_type, const float *pOrigin, edict_t *ed );
	void		(*pfnWriteByte)( int iValue );
	void		(*pfnWriteChar)( int iValue );
	void		(*pfnWriteShort)( int iValue );
	void		(*pfnWriteLong)( int iValue );
	void		(*pfnWriteAngle)( float flValue );
	void		(*pfnWriteCoord)( float flValue );
	void		(*pfnWriteString)( const char *sz );
	void		(*pfnWriteEntity)( int iValue );
	void		(*pfnMessageEnd)( void );

	// CRC routines
	void		(*pfnCRC32_Init)( dword *pulCRC );
	void		(*pfnCRC32_ProcessBuffer)( dword *pulCRC, void *p, int len );
	void		(*pfnCRC32_ProcessByte)( dword *pulCRC, byte ch );
	void		(*pfnCRC32_Final)( dword *pulCRC );
} sv_enginefuncs_t;

typedef struct serverfuncs_s
{
	size_t		api_size;	// sentinel to compare interface sizes for detect mismatch version

	// initialize/shutdown the game (one-time call after loading of game .dll )
	void		(*pfnGameInit)( void );
	void		(*pfnGameShutdown)( void );

	// save\restore system
	void		(*pfnSave)( edict_t *e, SAVERESTOREDATA *pSaveData );
	void		(*pfnSaveWriteFields)( SAVERESTOREDATA *pSaveData, const char *pname, void *pData, TYPEDESCRIPTION *pFileds, int count );
	void		(*pfnSaveGlobalState)( SAVERESTOREDATA *pSaveData );
	int 		(*pfnRestore)( edict_t *e, SAVERESTOREDATA *pSaveData, qboolean globalEntity );
	void		(*pfnSaveReadFields)( SAVERESTOREDATA *pSaveData, const char *pname, void *pData, TYPEDESCRIPTION *pField, int count );
	void		(*pfnRestoreGlobalState)( SAVERESTOREDATA *pSaveData );
	void		(*pfnSaveGameComment)( char *buffer, int max_length ); // build an actual save comment

	// changelevel callbacks
	void		(*pfnBuildAdjacentMapList)( SAVERESTOREDATA *pSaveData );
	void		(*pfnCreateEntitiesInTransitionList)( SAVERESTOREDATA *pSaveData, int levelMask );
	void		(*pfnCreateEntitiesInRestoreList)( SAVERESTOREDATA *pSaveData, const char *mapname );
	void		(*pfnResetGlobalState)( void );

	// new level callbacks
	void		(*pfnServerActivate)( edict_t *pEdictList, int edictCount, int clientMax );
	qboolean		(*pfnSpawnEntities)( char const *pMapName, const char *pStartSpot, char const *pEntityString );
	void		(*pfnServerDeactivate)( void );

	// physics frame
	void		(*pfnPhysicsFrame)( void );

	// client callbacks
	qboolean		(*pfnClientConnect)( edict_t *client, const char *pszName, const char *pszAddress, char szRejectReason[128] );
	void		(*pfnClientUserInfoChanged)( edict_t *client, const char *netname, char *infobuffer );
	void		(*pfnClientCommand)( edict_t *client, const cmd_buffer_t *cmd );
	void		(*pfnClientPutInServer)( edict_t *client );
	void		(*pfnClientDisconnect)( edict_t *client );
	void		(*pfnClientKill)( edict_t *client );

	// player move (sorted by order of calling)
	void		(*pfnCmdStart)( const edict_t *client, const struct usercmd_s *cmd, unsigned int random_seed );
	void		(*pfnPlayerPreThink)( edict_t *client, float frametime, double time );
	void		(*pfnPlayerMove)( struct playermove_s *ppmove, qboolean server );
	void		(*pfnPlayerTouch)( struct playermove_s *ppmove, edict_t *client ); // handle trigger touching for player
	void		(*pfnPlayerPostThink)( edict_t *client );
	void		(*pfnCmdEnd)( const edict_t *client );

	// snapshot callbacks
	void		(*pfnSetupVisibility)( edict_t *pViewEntity, edict_t *pClient, byte **pvs, byte **pas );
	qboolean		(*pfnAddToFullPack)( struct entity_state_s *state, int e, edict_t *ent, edict_t *host, int hostflags, qboolean player, byte *pSet );
	void		(*pfnUpdateClientData)( const edict_t *ent, int sendweapons, struct clientdata_s *cd );
	int		(*pfnGetWeaponData)( edict_t *player, struct weapon_data_s *info );

	// network initialization
	void		(*pfnCreateBaseline)( int player, int eindex, struct entity_state_s *baseline, edict_t *entity, int playermodelindex, const float *player_mins, const float *player_maxs );
	void		(*pfnRegisterEncoders)( void );

	// Return 1 if the packet is valid.  Set response_buffer_size if you want to send a response packet.  Incoming, it holds the max
	// size of the response_buffer, so you must zero it out if you choose not to respond.
	int		(*pfnConnectionlessPacket )( const struct netadr_s *net_from, const char *args, char *response_buffer, int *response_buffer_size );

	// One of the pfnForceUnmodified files failed the consistency check for the specified player
	// Return 0 to allow the client to continue, 1 to force immediate disconnection ( with an optional disconnect message of up to 256 characters )
	int		(*pfnInconsistentFile)( const struct edict_s *player, const char *filename, char *disconnect_message );

	// custom trace implementation (usually for SOLID_CUSTOM)
	void		(*pfnCalcAbsolutePosition)( edict_t *ent );
	void		(*pfnClipMoveToEntity)( edict_t *ent, const float *p0, const float *mins, const float *maxs, const float *p1, trace_t *tr );
	void		(*pfnClipPMoveToEntity)( struct physent_s *pe, const float *p0, const float *mins, const float *maxs, const float *p1, trace_t *tr );
	qboolean		(*pfnGetHullBounds)( int hullnumber, float *mins, float *maxs );
	qboolean		(*pfnShouldCollide)( edict_t *p1, edict_t *p2 );

	// optional callbacks (may be missed)
	void		(*pfnDrawDebugTriangles)( void );	// used for draw debug collisions for custom physic engine etc
	void		(*pfnDrawNormalTriangles)( void );	// used for draw debug overlay (textured)
	void		(*pfnDrawOrthoTriangles)( void );	// used for draw debug messages (2d mode)
	const char*	(*pfnGetGameDescription)( void );
	void		(*pfnCvarValue)( const edict_t *client, int requestID, const char *cvarName, const char *value );
	void		(*pfnOnFreeEntPrivateData)( edict_t *pEnt );
	int		(*pfnAllowSaveGame)( void );		// the game .dll should return 1 if save game should be allowed
	const char*	(*pfnGetString)( string_t iString );	// FIXME: remove this when interface will be finalized
} serverfuncs_t;

typedef int (*SERVERAPI)( int interfaceVersion, serverfuncs_t *pFunctionTable, sv_enginefuncs_t *engfuncs, struct globalvars_s *pGlobals );

#endif//SERVER_INT_H